/*
    Copyright 2004,2005,2006,2007,2009,2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <pthread.h>
#endif

#ifdef WIN32
    #define quick_thread(NAME, ARG) DWORD WINAPI NAME(ARG)
    #define thread_id   DWORD
#else
    #define quick_thread(NAME, ARG) void *NAME(ARG)
    #define thread_id   pthread_t
#endif

thread_id quick_threadx(void *func, void *data) {
    thread_id       tid;
#ifdef WIN32
    if(!CreateThread(NULL, 0, func, data, 0, &tid)) return(0);
#else
    pthread_attr_t  attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if(pthread_create(&tid, &attr, func, data)) return(0);
#endif
    return(tid);
}

typedef unsigned char   u8;
typedef unsigned short  u16;



#define VER     "0.2"
#define PORT    80
#define HTTP    "HTTP/1.1 200\r\n" \
                "Content-Length: %u\r\n" \
                "Content-Type: application/octet-stream\r\n" \
                "Connection: close\r\n" \
                "\r\n"



quick_thread(client, int sock);
int timeout(int sock, int secs);
void std_err(void);



u8      *filename   = NULL;
int     do_headers  = 1;



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    int     i,
            sd,
            sa,
            psz,
            on      = 1;
    u16     port    = PORT;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    fputs("\n"
        "One file only web server "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <file>\n"
            "\n"
            "Options:\n"
            "-p PORT  port to bind (%d)\n"
            "-x       the input file will be sent as is without the additional HTTP headers\n"
            "         useful for testing the own custom headers\n"
            "\n", argv[0], port);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            printf("\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case 'p': port          = atoi(argv[++i]);  break;
            case 'x': do_headers    = 0;                break;
            default: {
                printf("\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }
    filename = argv[argc];

    peer.sin_addr.s_addr = INADDR_ANY;
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;
    psz                  = sizeof(struct sockaddr_in);

    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))
      < 0) std_err();
    if(bind(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
      < 0) std_err();
    if(listen(sd, SOMAXCONN)
      < 0) std_err();

    printf("\n- waiting connections on port %d:\n\n", port);

    for(;;) {
        sa = accept(sd, (struct sockaddr *)&peer, &psz);
        if(sa < 0) std_err();

        printf("%s : %hu\n",
            inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

        if(!quick_threadx(client, (void *)sa)) close(sa);
    }

    close(sd);
    return(0);
}



quick_thread(client, int sock) {
    struct  stat    xstat;
    FILE    *fd     = NULL;
    int     t,
            len     = 0,
            buffsz  = 0;
    u8      *buff   = NULL,
            *p;

    do {
        if(buffsz <= len) { // so in any case buff is allocated with at least 1024 bytes
            buffsz += 1024;
            buff = realloc(buff, buffsz + 1);
            if(!buff) goto quit;
        }
        if(timeout(sock, 10) < 0) goto quit;
        t = recv(sock, buff + len, buffsz - len, 0);
        if(t <= 0) goto quit;
        len += t;
        buff[len] = 0;
    } while(!strstr(buff, "\r\n\r\n") && !strstr(buff, "\n\n"));

    for(p = buff; *p && (*p != '\r') && (*p != '\n'); p++);
    *p = 0;
    printf("  %s\n", buff);

    fd = fopen(filename, "rb"); // reload the file all the times, this is a test tool
    if(!fd) std_err();

    if(do_headers) {
        fstat(fileno(fd), &xstat);
        len = sprintf(buff, HTTP, (int)xstat.st_size);
        if(send(sock, buff, len, 0) != len) goto quit;
    }

    while((len = fread(buff, 1, buffsz, fd))) {
        if(send(sock, buff, len, 0) != len) goto quit;
    }

quit:
    if(fd) fclose(fd);
    if(buff) free(buff);
    //shutdown(sock, 2);
    close(sock);
    return(0);
}



int timeout(int sock, int secs) {
    struct  timeval tout;
    fd_set  fd_read;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    if(select(sock + 1, &fd_read, NULL, NULL, &tout)
      <= 0) return(-1);
    return(0);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


